package testcase;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.lang.reflect.InvocationTargetException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.text.DateFormat;
import java.util.Date;

import javax.swing.JFrame;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;

/**
 * Implementao de objeto remoto que  exportado no RMIRegistry para receber
 * chamadas RMI. As chamadas recebidas nas threads do RMI lanam a exceo
 * NullPointerException nas chamadas a mtodos da SwingUtilities devido ao
 * AppContext estar null quando a aplicao  executada pelo JavaWebStart. Esse
 * comportamento no  o esperado e passou a acontecer devido ao bug introduzido
 * no update25 de Java7.
 */
public class RmiServiceImpl extends java.rmi.server.UnicastRemoteObject
  implements RmiService {

  /**
   * Porta onde o RMI Registry executa por default
   */
  final static int PORT = 1099;

  /**
   * Uma janela Swing para exemplo.
   */
  static JFrame frame;
  /**
   * Um campo de texto Swing para exemplo.
   */
  static JTextField textField;
  /**
   * 
   */
  static PrintStream OUTPUT_FILE;

  /**
   * Construtor.
   * 
   * @throws RemoteException caso ocorra erro na criao do objeto remoto
   */
  public RmiServiceImpl() throws RemoteException {
    super();
  }

  /**
   * Programa principal que cria o objeto remoto e o exporta no RMIRegistry.
   * 
   * @param args the command line arguments
   * @throws Exception caso um erro ocorra durante a execuo do programa
   */
  public static void main(String[] args) throws Exception {
    File outputFile = new File("SAIDA.OUT");
    if (outputFile.exists()) {
      outputFile.delete();
    }
    OUTPUT_FILE = new PrintStream(new FileOutputStream(outputFile));
    OUTPUT_FILE.println(">> Criado em "
      + DateFormat.getInstance().format(new Date()));
    Registry reg;
    RmiServiceImpl service = new RmiServiceImpl();
    try {
      reg = LocateRegistry.getRegistry(PORT);
      reg.rebind("demo", service);
    }
    catch (RemoteException ex) {
      reg = LocateRegistry.createRegistry(PORT);
      reg.rebind("demo", service);
    }
    SwingThreadDispatcher.setPrinter(OUTPUT_FILE, System.out);
    SwingThreadDispatcher.init();
    SwingUtilities.invokeAndWait(new Runnable() {
      @Override
      public void run() {
        frame = new JFrame("Janela de Teste");
        textField =
          new JTextField("Antes da chamada de callback na thread do RMI");
        frame.getContentPane().add(textField);
        frame.pack();
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        frame.setVisible(true);
      }
    });
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public void testInvokeLater(boolean useWorkAround) {
    showMessage(OUTPUT_FILE, ">> testInvokeLater");
    showMessage(System.out, ">> testInvokeLater");
    Runnable rn = new Runnable() {
      @Override
      public void run() {
        textField.setText("testInvokeLater executado com sucesso.");
        frame.pack();
      }
    };
    try {
      if (!useWorkAround) {
        SwingUtilities.invokeLater(rn);
      }
      else {
        SwingThreadDispatcher.invokeLater(rn);
      }
    }
    catch (NullPointerException e) {
      String msg = ">> Erro na execucao de testInvokeLater";
      showMessage(OUTPUT_FILE, msg, e);
      showMessage(System.out, msg, e);

    }
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public void testIsEventDispatchThread(boolean useWorkAround) {
    showMessage(OUTPUT_FILE, ">> testIsEventDispatchThread");
    showMessage(System.out, ">> testIsEventDispatchThread");
    Runnable rn = new Runnable() {
      @Override
      public void run() {
        textField.setText("testIsEventDispatchThread executado com sucesso.");
        frame.pack();
      }
    };
    try {
      if (!useWorkAround) {
        if (SwingUtilities.isEventDispatchThread()) {
          rn.run();
        }
        else {
          SwingUtilities.invokeLater(rn);
        }
      }
      else {
        if (SwingThreadDispatcher.isEventDispatchThread()) {
          rn.run();
        }
        else {
          SwingThreadDispatcher.invokeLater(rn);
        }
      }
    }
    catch (NullPointerException e) {
      String msg = ">> Erro na execucao de testIsEventDispatchThread";
      showMessage(OUTPUT_FILE, msg, e);
      showMessage(System.out, msg, e);
    }
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public void testInvokeAndWait(boolean useWorkAround) {
    showMessage(OUTPUT_FILE, ">> testInvokeAndWait");
    showMessage(System.out, ">> testInvokeAndWait");
    Runnable rn = new Runnable() {
      @Override
      public void run() {
        textField.setText("testInvokeAndWait executado com sucesso.");
        frame.pack();
      }
    };
    try {
      if (!useWorkAround) {
        SwingUtilities.invokeAndWait(rn);
      }
      else {
        SwingThreadDispatcher.invokeAndWait(rn);
      }
    }
    catch (InterruptedException e) {
      e.printStackTrace();
    }
    catch (InvocationTargetException e) {
      e.printStackTrace();
    }
    catch (NullPointerException e) {
      String msg = ">> Erro na execucao de testInvokeAndWait";
      showMessage(OUTPUT_FILE, msg, e);
      showMessage(System.out, msg, e);
    }
  }

  /**
   * Mostra uma mensagem no output stream.
   * 
   * @param stream o stream de output da mensagem
   * @param msg a mensagem
   * @param e a exceo, quando houver
   */
  private void showMessage(PrintStream stream, String msg, Exception e) {
    stream.println(msg);
    if (e != null) {
      e.printStackTrace(stream);
    }
    stream.flush();
  }

  /**
   * Mostra uma mensagem no output stream.
   * 
   * @param stream o stream de output da mensagem
   * @param msg a mensagem
   */
  private void showMessage(PrintStream stream, String msg) {
    showMessage(stream, msg, null);
  }
}